﻿using Microsoft.Win32;
using System;
using System.Diagnostics;
using System.Drawing;
using System.IO;
using System.Reflection;
using System.Runtime.InteropServices;
using System.Runtime.InteropServices.ComTypes;
using System.Text;
using OCR.TesseractWrapper;

namespace IFilterPicture
{
    /// <summary>
    /// The IFilter and IPersistFile interfaces, and some support structure
    /// for the File Filter (Extractor)
    /// </summary>
    [ComVisible(true), Guid("78BD7653-D16A-43ad-B119-7080D635F27E")]
    public class IFilterPicture : IPersistFile, IFilter
    {
        private static string myGuid = "78BD7653-D16A-43ad-B119-7080D635F27E"; /// <summary>For convenience, GUID in a string</summary>

        private bool bApplyIndexAttributes = false;
        private bool bCrawlAttributes = false;
        private bool bApplyOther = false;

        private bool doneFirstChunk = false;

        private string currentFileName = string.Empty;
        private int iChunkCount = 0;

        private const uint LANGUAGE_FR = 0x040C;

        private string text;

        public IFilterPicture()
        {
        }

        //	IFilter Methods
        /// <summary>Initializes a filtering session.</summary>
        /// <param name="grfFlags">What processing to perform</param>
        /// <param name="cAttributes">How many Attribute elements are in the aAttributes array, may be 0</param>
        /// <param name="aAttributes">Attributes to process</param>
        /// <param name="pdwFlags"></param>
        public void Init([MarshalAs(UnmanagedType.U4)] IFILTER_INIT grfFlags, uint cAttributes, [MarshalAs(UnmanagedType.LPArray, SizeParamIndex = 1)] FULLPROPSPEC[] aAttributes, ref uint pdwFlags)
        {
            bApplyIndexAttributes = ((grfFlags & IFILTER_INIT.APPLY_INDEX_ATTRIBUTES) == IFILTER_INIT.APPLY_INDEX_ATTRIBUTES);
            bCrawlAttributes = ((grfFlags & IFILTER_INIT.APPLY_CRAWL_ATTRIBUTES) == IFILTER_INIT.APPLY_CRAWL_ATTRIBUTES);
            bApplyOther = ((grfFlags & IFILTER_INIT.APPLY_OTHER_ATTRIBUTES) == IFILTER_INIT.APPLY_OTHER_ATTRIBUTES);
        }

        /// <summary>
        /// Loads the first chunk or sets the pointers to zero on fail
        /// </summary>
        /// <returns></returns>
        private bool getFirstChunk()
        {
            if ((currentFileName == null) || (File.Exists(currentFileName) == false))
            {
                currentFileName = "";
                return false;
            }

            try
            {
                TesseractProcessor p = new TesseractProcessor();
                if (p.Init(Path.GetDirectoryName(Assembly.GetAssembly(typeof(IFilterPicture)).Location) + @"\fra.traineddata", "fra", (int)eOcrEngineMode.OEM_DEFAULT))
                {
                    text = p.Recognize(currentFileName);
                }
                else
                {
                    text = string.Empty;
                }
            }
            catch (Exception e)
            {
                text = string.Empty;
            }

            return true;
        }

        /// <summary>The IFilter::GetChunk method positions the filter at the beginning of the next chunk, or at the first chunk if this is the first call to the GetChunk method, and returns a description of the current chunk.</summary>
        /// <remarks>Each chunk might be tiny; the file extractor (filter) decides.
        /// Note that Chunk ID 0 is invalid; must start at 1.</remarks>
        /// <param name="pStat">STAT_CHUNK pointer</param>
        [PreserveSig]
        public int GetChunk(out STAT_CHUNK pStat)
        {
            bool bSuccess = true;
            STAT_CHUNK myStat = new STAT_CHUNK();
            pStat = myStat;
            pStat.locale = LANGUAGE_FR;
            iChunkCount++;
            pStat.idChunk = (uint)iChunkCount;

            if (iChunkCount == 2)
            {
                return Constants.FILTER_E_END_OF_CHUNKS;
            }

            if (!doneFirstChunk)
            {
                doneFirstChunk = true;
                bSuccess = getFirstChunk();
            }

            if (!bSuccess)
            {
                return Constants.FILTER_E_END_OF_CHUNKS;
            }
            pStat.flags = CHUNKSTATE.CHUNK_TEXT;

            pStat.attribute.guidPropSet = new Guid(0xb725f130, 0x47ef, 0x101a, 0xa5, 0xf1, 0x02, 0x60, 0x8c, 0x9e, 0xeb, 0xac);
            pStat.attribute.psProperty.ulKind = (int)PSKIND.PROPID;
            pStat.attribute.psProperty.propid = Constants.PID_STG_CONTENTS;
            pStat.idChunkSource = pStat.idChunk;
            pStat.breakType = CHUNK_BREAKTYPE.CHUNK_EOC;
            pStat.cwcStartSource = 0;
            return Constants.S_OK;
        }

        /// <summary>Gets the text from this chunk of type CHUNK_TEXT</summary>
        /// <param name="pcwcBuffer">Size of input buffer in UNICODE characters; fill with output size in CHARACTERS</param>
        /// <param name="buffer">Null-terminated UNICODE string of the text.</param>
        /// <returns></returns>
        /// <remarks>This is called repeatedly INSIDE THE ONE CHUNK until all the text 
        /// is delivered.</remarks>
        [PreserveSig]
        public int GetText(ref int pcwcBuffer, IntPtr buffer)
        {
            pcwcBuffer = text.Length;
            Marshal.Copy(text.ToCharArray(), 0, buffer, text.Length);
            return Constants.FILTER_S_LAST_TEXT;
        }

        /// <summary>Retrieves a value (internal value-type property) from a chunk, which must have a CHUNKSTATE enumeration value of CHUNK_VALUE.</summary>
        /// <param name="ppPropValue"></param>
        /// <remarks>According to the blog, the most important ones are:
        /// DocAuthor, PrimaryDate, PercievedType and DocTitle.
        /// PerceivedTypes include contact, communications/calendar
        ///		document/note and communications/task</remarks>
        ///		
        [PreserveSig]
        public int GetValue(UIntPtr ppPropValue)
        {
            return Constants.FILTER_E_NO_VALUES;
        }

        /// <summary>Returns the class identifier (CLSID) for the component object. In the case 
        /// of a filter, this is the file type identifier.</summary>
        /// <param name="pClassID"></param>
        public void GetClassID(out Guid pClassID)
        {
            pClassID = new Guid(myGuid);
        }

        public void BindRegion([MarshalAs(UnmanagedType.Struct)]FILTERREGION origPos, ref Guid riid, ref UIntPtr ppunk)
        {
            return;
        }


        /// <summary>Open and initialize the file</summary>
        /// <param name="pszFileName">Absolute file path to open/initialize</param>
        /// <param name="dwMode">STGM flags for how to open</param>
        /// <remarks>The STGM Flags provide 18 options and additional combinations for how to open
        /// and process the file.  Since IFilters are inherently read-only to the IPersist:: file, 
        /// most of these are irrelevant.  So we're ignoring it.</remarks>
        public void Load(string pszFileName, int dwMode)
        {
            currentFileName = pszFileName;
        }

        //	These are stubs needed for the interface but never called for Read-Only Access
        public void GetCurFile(out string ppszFileName) { ppszFileName = currentFileName; }
        public int IsDirty() { return Constants.E_NOTIMPL; }
        public void Save(string pszFileName, bool fRemember) { }
        public void SaveCompleted(string pszFileName) { }
    }

    [Flags]
    public enum IFILTER_INIT
    {
        NONE = 0,
        CANON_PARAGRAPHS = 1,
        HARD_LINE_BREAKS = 2,
        CANON_HYPHENS = 4,
        CANON_SPACES = 8,
        APPLY_INDEX_ATTRIBUTES = 16,
        APPLY_CRAWL_ATTRIBUTES = 256,
        APPLY_OTHER_ATTRIBUTES = 32,
        INDEXING_ONLY = 64,
        SEARCH_LINKS = 128,
        FILTER_OWNED_VALUE_OK = 512
    }

    [Flags]
    public enum IFILTER_FLAGS
    {
        OLE_PROPERTIES = 1
    }

    /// <summary>Describes what separates the CURRENT chunk from the PREVIOUS chunk with the same Attribute.
    /// Used by IFilter::GetChunk() in the return.</summary>
    public enum CHUNK_BREAKTYPE
    {
        CHUNK_NO_BREAK = 0,	/// <summary>No Break from previous and doesn't get flags/attributes; glue them </summary>
        CHUNK_EOW = 1,		/// <summary>Reached the End-Of-Word.  Don't use it; language dependent.</summary>
        CHUNK_EOS = 2,		/// <summary>Sentence Break</summary>
        CHUNK_EOP = 3,		///	<summary>Paragraph Break</summary>
        CHUNK_EOC = 4		///	<summary>Chapter Break</summary>
    }

    [Flags]
    public enum CHUNKSTATE
    {
        CHUNK_TEXT = 0x1,
        CHUNK_VALUE = 0x2,
        CHUNK_FILTER_OWNED_VALUE = 0x4
    }

    public enum PSKIND
    {
        LPWSTR = 0,
        PROPID = 1
    }

    // [StructLayout(LayoutKind.Sequential)]
    [StructLayout(LayoutKind.Explicit)]
    public struct PROPSPEC
    {
        [FieldOffset(0)]
        public uint ulKind;
        [FieldOffset(4)]
        public uint propid;

        //	NOTE: The real structure uses a Union.  C# doesn't support unions.  So I've commented
        //		  out this element to make the correct size.  Other samples don't do this.  TMcN
        [FieldOffset(4)]
        public IntPtr lpwstr;
    }

    [StructLayout(LayoutKind.Sequential)]
    public struct FULLPROPSPEC
    {
        public Guid guidPropSet;
        public PROPSPEC psProperty;
    }

    [StructLayout(LayoutKind.Sequential)]
    public struct STAT_CHUNK
    {
        public uint idChunk;
        [MarshalAs(UnmanagedType.U4)]
        public CHUNK_BREAKTYPE breakType;
        [MarshalAs(UnmanagedType.U4)]
        public CHUNKSTATE flags;
        public uint locale;
        [MarshalAs(UnmanagedType.Struct)]
        public FULLPROPSPEC attribute;
        public uint idChunkSource;
        public uint cwcStartSource;
        public uint cwcLenSource;
    }

    [StructLayout(LayoutKind.Sequential)]
    public struct FILTERREGION
    {
        public uint idChunk;
        public uint cwcStart;
        public uint cwcExtent;
    }


    [StructLayout(LayoutKind.Sequential)]
    public struct PROPVARIANT
    {
        public short vt;			/// <summary>Variant Type</summary>
        public short reserved1;
        public short reserved2;
        public short reserved3;
        unsafe
        public char* pwszVal;
    }


    [ComImport]
    [Guid("89BCB740-6119-101A-BCB7-00DD010655AF")]
    [InterfaceType(ComInterfaceType.InterfaceIsIUnknown)]
    public interface IFilter
    {
        void Init([MarshalAs(UnmanagedType.U4)] IFILTER_INIT grfFlags, uint cAttributes,
            [MarshalAs(UnmanagedType.LPArray, SizeParamIndex = 1)] FULLPROPSPEC[] aAttributes,
            ref uint pdwFlags);
        [PreserveSig]
        int GetChunk(out STAT_CHUNK pStat);


        [PreserveSig]
        int GetText(ref int pcwcBuffer, IntPtr buffer);
        [PreserveSig]
        int GetValue(UIntPtr ppPropValue);

        void BindRegion([MarshalAs(UnmanagedType.Struct)]FILTERREGION origPos, ref Guid riid, ref UIntPtr ppunk);
    }

    [ComImport]
    [Guid("f07f3920-7b8c-11cf-9be8-00aa004b9986")]
    public class CFilter
    {
    }

    public class Constants
    {
        public const uint PID_STG_DIRECTORY = 0x00000002;
        public const uint PID_STG_CLASSID = 0x00000003;
        public const uint PID_STG_STORAGETYPE = 0x00000004;
        public const uint PID_STG_VOLUME_ID = 0x00000005;
        public const uint PID_STG_PARENT_WORKID = 0x00000006;
        public const uint PID_STG_SECONDARYSTORE = 0x00000007;
        public const uint PID_STG_FILEINDEX = 0x00000008;
        public const uint PID_STG_LASTCHANGEUSN = 0x00000009;
        public const uint PID_STG_NAME = 0x0000000a;
        public const uint PID_STG_PATH = 0x0000000b;
        public const uint PID_STG_SIZE = 0x0000000c;
        public const uint PID_STG_ATTRIBUTES = 0x0000000d;
        public const uint PID_STG_WRITETIME = 0x0000000e;
        public const uint PID_STG_CREATETIME = 0x0000000f;
        public const uint PID_STG_ACCESSTIME = 0x00000010;
        public const uint PID_STG_CHANGETIME = 0x00000011;
        public const uint PID_STG_CONTENTS = 0x00000013;
        public const uint PID_STG_SHORTNAME = 0x00000014;
        public const int FILTER_E_END_OF_CHUNKS = (unchecked((int)0x80041700));
        public const int FILTER_E_NO_MORE_TEXT = (unchecked((int)0x80041701));
        public const int FILTER_E_NO_MORE_VALUES = (unchecked((int)0x80041702));
        public const int FILTER_E_NO_TEXT = (unchecked((int)0x80041705));
        public const int FILTER_E_NO_VALUES = (unchecked((int)0x80041706));
        public const int FILTER_S_LAST_TEXT = (unchecked((int)0x00041709));
        public const int S_OK = (unchecked((int)0x00000000L));
        public const int S_FALSE = (unchecked((int)0x00000001L));
        public const int E_NOTIMPL = (unchecked((int)0x80004001L));
        public const int E_ABORT = (unchecked((int)0x80004004L));
    }

    /*
    public class TesseractOCR
    {
        private string commandpath;
        private string outpath;

        public TesseractOCR(string commandpath)
        {
            this.commandpath = commandpath;
            outpath = Path.ChangeExtension(Path.GetTempFileName(), ".txt");
        }

        private string analyze(string filename)
        {
            string args = string.Format("\"{0}\" \"{1}\" -l fra", filename, outpath.Replace(".txt", string.Empty));
            ProcessStartInfo startinfo = new ProcessStartInfo(commandpath, args);
            startinfo.CreateNoWindow = true;
            startinfo.UseShellExecute = false;

            Process.Start(startinfo).WaitForExit();

            string ret = "";
            using (StreamReader r = new StreamReader(outpath))
            {
                string content = r.ReadToEnd();
                ret = content;
            }
            File.Delete(outpath);
            return ret;
        }

        public string OCRFromFile(string filename)
        {
            return analyze(filename);
        }
    }
    */
}
